"""Define package errors."""
from __future__ import annotations

from typing import Any


class OpenUvError(Exception):
    """Define a base error."""

    pass


class ApiUnavailableError(OpenUvError):
    """Define an error related to the OpenUV API being down."""

    pass


class InvalidApiKeyError(OpenUvError):
    """Define an error related to invalid API keys."""

    pass


class RateLimitExceededError(OpenUvError):
    """Define an error related to exceeding the daily rate limit."""

    pass


class RequestError(OpenUvError):
    """Define an error related to invalid requests."""

    pass


ERROR_MESSAGE_MAP = {
    "User with API Key not found": InvalidApiKeyError,
    "Daily API quota exceeded": RateLimitExceededError,
}


def raise_error(
    endpoint: str, payload: dict[str, Any], raising_err: Exception | None
) -> None:
    """Raise the appropriate error based on the response data.

    Args:
        endpoint: The endpoint that was queried.
        payload: An API response payload.
        raising_err: The exception that caused the overall issue.

    Raises:
        exc: Raised upon an HTTP error.
    """
    if (error_msg := payload.get("error")) is None:
        return

    try:
        exc = next(exc for idx, exc in ERROR_MESSAGE_MAP.items() if idx in error_msg)
    except StopIteration:
        exc = RequestError

    exc.__cause__ = raising_err
    raise exc(f"Error while querying {endpoint}: {error_msg}")



"""Define dynamic text fixtures."""
from __future__ import annotations

import json
from typing import Any, cast

import pytest

from .common import load_fixture


@pytest.fixture(name="api_statistics_response", scope="session")
def api_statistics_response_fixture() -> dict[str, Any]:
    """Define a fixture to return an API status response.

    Returns:
        An API response payload.
    """
    return cast(
        dict[str, Any], json.loads(load_fixture("api_statistics_response.json"))
    )


@pytest.fixture(name="api_status_response", scope="session")
def api_status_response_fixture() -> dict[str, Any]:
    """Define a fixture to return an API status response.

    Returns:
        An API response payload.
    """
    return cast(dict[str, Any], json.loads(load_fixture("api_status_response.json")))


@pytest.fixture(name="error_invalid_api_key_response", scope="session")
def error_invalid_api_key_response_fixture() -> dict[str, Any]:
    """Define a fixture to return an invalid API key error response.

    Returns:
        An API response payload.
    """
    return cast(
        dict[str, Any], json.loads(load_fixture("error_invalid_api_key_response.json"))
    )


@pytest.fixture(name="error_rate_limit_response", scope="session")
def error_rate_limit_response_fixture() -> dict[str, Any]:
    """Define a fixture to return an rate limit error response.

    Returns:
        An API response payload.
    """
    return cast(
        dict[str, Any], json.loads(load_fixture("error_rate_limit_response.json"))
    )


@pytest.fixture(name="protection_window_response", scope="session")
def protection_window_response_fixture() -> dict[str, Any]:
    """Define a fixture to return a protection window response.

    Returns:
        An API response payload.
    """
    return cast(
        dict[str, Any], json.loads(load_fixture("protection_window_response.json"))
    )


@pytest.fixture(name="uv_forecast_response", scope="session")
def uv_forecast_response_fixture() -> dict[str, Any]:
    """Define a fixture to return an UV forecast response.

    Returns:
        An API response payload.
    """
    return cast(dict[str, Any], json.loads(load_fixture("uv_forecast_response.json")))


@pytest.fixture(name="uv_index_response", scope="session")
def uv_index_response_fixture() -> dict[str, Any]:
    """Define a fixture to return an UV index response.

    Returns:
        An API response payload.
    """
    return cast(dict[str, Any], json.loads(load_fixture("uv_index_response.json")))



"""Define common test utilities."""
from __future__ import annotations

import os

TEST_ALTITUDE = 1609.3
TEST_API_KEY = "12345"
TEST_LATITUDE = 51.528308
TEST_LONGITUDE = -0.3817803


def load_fixture(filename: str) -> str:
    """Load a fixture.

    Args:
        filename: The filename of the fixtures/ file to load.

    Returns:
        A string containing the contents of the file.
    """
    path = os.path.join(os.path.dirname(__file__), "fixtures", filename)
    with open(path, encoding="utf-8") as fptr:
        return fptr.read()



"""Run an example script to quickly test."""
import asyncio
import logging
import time

from aiohttp import ClientSession

from pyopenuv import Client
from pyopenuv.errors import OpenUvError

_LOGGER = logging.getLogger(__name__)

API_KEY = "<OPENUV_API_KEY>"
LATITUDE = 39.7974509
LONGITUDE = -104.8887227
ALTITUDE = 1609.3


async def main() -> None:
    """Create the aiohttp session and run the example."""
    logging.basicConfig(level=logging.DEBUG)

    async with ClientSession() as session:
        client = Client(
            API_KEY,
            LATITUDE,
            LONGITUDE,
            altitude=ALTITUDE,
            session=session,
        )

        start = time.time()

        try:
            # Get current UV info:
            _LOGGER.info("CURRENT UV DATA:")
            _LOGGER.info(await client.uv_index())

            # Get forecasted UV info:
            _LOGGER.info("FORECASTED UV DATA:")
            _LOGGER.info(await client.uv_forecast())

            # Get UV protection window:
            _LOGGER.info("UV PROTECTION WINDOW:")
            _LOGGER.info(await client.uv_protection_window())
        except OpenUvError as err:
            _LOGGER.info(err)

    end = time.time()

    _LOGGER.info("Execution time: %s seconds", end - start)


asyncio.run(main())



"""Define a client to interact with openuv.io."""
from __future__ import annotations

import asyncio
from typing import Any, cast

from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError, ContentTypeError

from .const import LOGGER
from .errors import ApiUnavailableError, raise_error

API_URL_SCAFFOLD = "https://api.openuv.io/api/v1"

DEFAULT_PROTECTION_HIGH = 3.5
DEFAULT_PROTECTION_LOW = 3.5
DEFAULT_TIMEOUT = 30


class Client:
    """Define the client."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        api_key: str,
        latitude: float,
        longitude: float,
        *,
        altitude: float = 0.0,
        session: ClientSession | None = None,
        check_status_before_request: bool = False,
    ) -> None:
        """Initialize.

        Args:
            api_key: An OpenUV API key.
            latitude: A latitude.
            longitude: A longitude.
            altitude: An altitude.
            session: An optional aiohttp ClientSession.
            check_status_before_request: Whether the API status should be checked prior
                to every request.
        """
        self._api_key = api_key
        self._session = session
        self.altitude = str(altitude)
        self.check_status_before_request = check_status_before_request
        self.latitude = str(latitude)
        self.longitude = str(longitude)

    async def _async_check_api_status_if_required(self) -> None:
        """Check the status of the API if configured to do so.

        Raises:
            ApiUnavailableError: Raised when the API is unavailable.
        """
        if not self.check_status_before_request or await self.api_status():
            return

        raise ApiUnavailableError("The OpenUV API is unavailable")

    async def _async_request(
        self, method: str, endpoint: str, **kwargs: dict[str, str]
    ) -> dict[str, Any]:
        """Make an API request.

        Args:
            method: An HTTP method.
            endpoint: A relative API endpoint.
            **kwargs: Additional kwargs to send with the request.

        Returns:
            An API response payload.
        """
        kwargs.setdefault("headers", {})
        kwargs["headers"]["x-access-token"] = self._api_key

        kwargs.setdefault("params", {})
        kwargs["params"]["lat"] = self.latitude
        kwargs["params"]["lng"] = self.longitude
        kwargs["params"]["alt"] = self.altitude

        if use_running_session := self._session and not self._session.closed:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        data: dict[str, Any] = {}
        url = f"{API_URL_SCAFFOLD}/{endpoint}"

        try:
            async with session.request(method, url, **kwargs) as resp:
                data = await resp.json()
                raising_err = None

                try:
                    resp.raise_for_status()
                except ClientError as err:
                    raising_err = err

                raise_error(endpoint, data, raising_err)
        except (asyncio.TimeoutError, ContentTypeError) as err:
            raise_error(endpoint, {"error": str(err)}, err)
        finally:
            if not use_running_session:
                await session.close()

        LOGGER.debug("Data received for %s: %s", endpoint, data)

        return data

    async def api_statistics(self) -> dict[str, Any]:
        """Get API usage statistics.

        Returns:
            An API response payload.
        """
        await self._async_check_api_status_if_required()
        return await self._async_request("get", "stat")

    async def api_status(self) -> bool:
        """Get the current status of the API.

        Returns:
            True if the API is available, False if it is unavailable
        """
        resp = await self._async_request("get", "status")
        return cast(bool, resp["status"])

    async def uv_forecast(self) -> dict[str, Any]:
        """Get forecasted UV data.

        Returns:
            An API response payload.
        """
        await self._async_check_api_status_if_required()
        return await self._async_request("get", "forecast")

    async def uv_index(self) -> dict[str, Any]:
        """Get current UV data.

        Returns:
            An API response payload.
        """
        await self._async_check_api_status_if_required()
        return await self._async_request("get", "uv")

    async def uv_protection_window(
        self, low: float = DEFAULT_PROTECTION_LOW, high: float = DEFAULT_PROTECTION_HIGH
    ) -> dict[str, Any]:
        """Get data on when a UV protection window is.

        Args:
            low: The low end of the UV index to monitor.
            high: The high end of the UV index to monitor.

        Returns:
            An API response payload.
        """
        await self._async_check_api_status_if_required()
        return await self._async_request(
            "get", "protection", params={"from": str(low), "to": str(high)}
        )



"""Define tests for the client object."""
from __future__ import annotations

import asyncio
from typing import Any
from unittest.mock import patch

import aiohttp
import pytest
from aresponses import ResponsesMockServer

from pyopenuv import Client
from pyopenuv.errors import (
    ApiUnavailableError,
    InvalidApiKeyError,
    RateLimitExceededError,
    RequestError,
)
from tests.common import TEST_ALTITUDE, TEST_API_KEY, TEST_LATITUDE, TEST_LONGITUDE


@pytest.mark.asyncio
async def test_api_statistics(
    aresponses: ResponsesMockServer, api_statistics_response: dict[str, Any]
) -> None:
    """Test successfully retrieving API usage statistics.

    Args:
        aresponses: An aresponses server.
        api_statistics_response: An API response payload.
    """
    aresponses.add(
        "api.openuv.io",
        "/api/v1/stat",
        "get",
        response=aiohttp.web_response.json_response(
            api_statistics_response, status=200
        ),
    )

    async with aiohttp.ClientSession() as session:
        client = Client(
            TEST_API_KEY,
            TEST_LATITUDE,
            TEST_LONGITUDE,
            altitude=TEST_ALTITUDE,
            session=session,
        )
        data = await client.api_statistics()
        assert data == {
            "result": {
                "requests_today": 25,
                "requests_yesterday": 14,
                "requests_month": 146,
                "requests_last_month": 446,
                "cost_today": 0,
                "cost_yesterday": 0,
                "cost_month": 0.024,
                "cost_last_month": 0,
            }
        }

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_bad_request(aresponses: ResponsesMockServer) -> None:
    """Test that the proper exception is raised during a bad request.

    Args:
        aresponses: An aresponses server.
    """
    aresponses.add(
        "api.openuv.io",
        "/api/v1/bad_endpoint",
        "get",
        aresponses.Response(text="", status=500),
    )

    with pytest.raises(RequestError):
        async with aiohttp.ClientSession() as session:
            client = Client(
                TEST_API_KEY,
                TEST_LATITUDE,
                TEST_LONGITUDE,
                altitude=TEST_ALTITUDE,
                session=session,
            )
            await client._async_request(  # pylint: disable=protected-access
                "get", "bad_endpoint"
            )

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_error_invalid_api_key(
    aresponses: ResponsesMockServer, error_invalid_api_key_response: dict[str, Any]
) -> None:
    """Test the that the proper exception is raised with a bad API key.

    Args:
        aresponses: An aresponses server.
        error_invalid_api_key_response: An API response payload.
    """
    aresponses.add(
        "api.openuv.io",
        "/api/v1/protection",
        "get",
        response=aiohttp.web_response.json_response(
            error_invalid_api_key_response, status=403
        ),
    )

    with pytest.raises(InvalidApiKeyError):
        async with aiohttp.ClientSession() as session:
            client = Client(
                TEST_API_KEY,
                TEST_LATITUDE,
                TEST_LONGITUDE,
                altitude=TEST_ALTITUDE,
                session=session,
            )
            await client.uv_protection_window()

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_error_rate_limited(
    aresponses: ResponsesMockServer, error_rate_limit_response: dict[str, Any]
) -> None:
    """Test the that the proper exception is raised when the rate limit is reached.

    Args:
        aresponses: An aresponses server.
        error_rate_limit_response: An API response payload.
    """
    aresponses.add(
        "api.openuv.io",
        "/api/v1/protection",
        "get",
        response=aiohttp.web_response.json_response(
            error_rate_limit_response, status=403
        ),
    )

    with pytest.raises(RateLimitExceededError):
        async with aiohttp.ClientSession() as session:
            client = Client(
                TEST_API_KEY,
                TEST_LATITUDE,
                TEST_LONGITUDE,
                altitude=TEST_ALTITUDE,
                session=session,
            )
            await client.uv_protection_window()

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_protection_window(
    aresponses: ResponsesMockServer, protection_window_response: dict[str, Any]
) -> None:
    """Test successfully retrieving the protection window.

    Args:
        aresponses: An aresponses server.
        protection_window_response: An API response payload.
    """
    aresponses.add(
        "api.openuv.io",
        "/api/v1/protection",
        "get",
        response=aiohttp.web_response.json_response(
            protection_window_response, status=200
        ),
    )

    async with aiohttp.ClientSession() as session:
        client = Client(
            TEST_API_KEY,
            TEST_LATITUDE,
            TEST_LONGITUDE,
            altitude=TEST_ALTITUDE,
            session=session,
        )
        data = await client.uv_protection_window()
        assert data == {
            "result": {
                "from_time": "2018-07-30T15:17:49.750Z",
                "from_uv": 3.2509,
                "to_time": "2018-07-30T22:47:49.750Z",
                "to_uv": 3.6483,
            }
        }

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_session_from_scratch(
    aresponses: ResponsesMockServer, uv_forecast_response: dict[str, Any]
) -> None:
    """Test that an aiohttp ClientSession is created on the fly if needed.

    Args:
        aresponses: An aresponses server.
        uv_forecast_response: An API response payload.
    """
    aresponses.add(
        "api.openuv.io",
        "/api/v1/forecast",
        "get",
        response=aiohttp.web_response.json_response(uv_forecast_response, status=200),
    )

    client = Client(TEST_API_KEY, TEST_LATITUDE, TEST_LONGITUDE, altitude=TEST_ALTITUDE)
    data = await client.uv_forecast()
    assert data == {
        "result": [
            {
                "uv": 0,
                "uv_time": "2018-07-30T11:57:49.750Z",
                "sun_position": {
                    "azimuth": -2.0081567900835937,
                    "altitude": -0.011856950133816461,
                },
            },
            {
                "uv": 0.2446,
                "uv_time": "2018-07-30T12:57:49.750Z",
                "sun_position": {
                    "azimuth": -1.845666592871966,
                    "altitude": 0.1764062658258758,
                },
            },
        ]
    }

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_timeout() -> None:
    """Test that a timeout raises an exception."""
    async with aiohttp.ClientSession() as session:
        client = Client(
            TEST_API_KEY,
            TEST_LATITUDE,
            TEST_LONGITUDE,
            altitude=TEST_ALTITUDE,
            session=session,
        )

        with patch(
            "aiohttp.ClientSession.request", side_effect=asyncio.TimeoutError
        ), pytest.raises(RequestError):
            await client.uv_forecast()


@pytest.mark.asyncio
async def test_uv_forecast(
    aresponses: ResponsesMockServer, uv_forecast_response: dict[str, Any]
) -> None:
    """Test successfully retrieving UV forecast info.

    Args:
        aresponses: An aresponses server.
        uv_forecast_response: An API response payload.
    """
    aresponses.add(
        "api.openuv.io",
        "/api/v1/forecast",
        "get",
        response=aiohttp.web_response.json_response(uv_forecast_response, status=200),
    )

    async with aiohttp.ClientSession() as session:
        client = Client(
            TEST_API_KEY,
            TEST_LATITUDE,
            TEST_LONGITUDE,
            altitude=TEST_ALTITUDE,
            session=session,
        )
        data = await client.uv_forecast()
        assert data == {
            "result": [
                {
                    "uv": 0,
                    "uv_time": "2018-07-30T11:57:49.750Z",
                    "sun_position": {
                        "azimuth": -2.0081567900835937,
                        "altitude": -0.011856950133816461,
                    },
                },
                {
                    "uv": 0.2446,
                    "uv_time": "2018-07-30T12:57:49.750Z",
                    "sun_position": {
                        "azimuth": -1.845666592871966,
                        "altitude": 0.1764062658258758,
                    },
                },
            ]
        }

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_uv_index(
    aresponses: ResponsesMockServer, uv_index_response: dict[str, Any]
) -> None:
    """Test successfully retrieving UV index info.

    Args:
        aresponses: An aresponses server.
        uv_index_response: An API response payload.
    """
    aresponses.add(
        "api.openuv.io",
        "/api/v1/uv",
        "get",
        response=aiohttp.web_response.json_response(uv_index_response, status=200),
    )

    async with aiohttp.ClientSession() as session:
        client = Client(
            TEST_API_KEY,
            TEST_LATITUDE,
            TEST_LONGITUDE,
            altitude=TEST_ALTITUDE,
            session=session,
        )
        data = await client.uv_index()
        assert data == {
            "result": {
                "uv": 8.2342,
                "uv_time": "2018-07-30T20:53:06.302Z",
                "uv_max": 10.3335,
                "uv_max_time": "2018-07-30T19:07:11.505Z",
                "ozone": 300.7,
                "ozone_time": "2018-07-30T18:07:04.466Z",
                "safe_exposure_time": {
                    "st1": 20,
                    "st2": 24,
                    "st3": 32,
                    "st4": 40,
                    "st5": 65,
                    "st6": 121,
                },
                "sun_info": {
                    "sun_times": {
                        "solarNoon": "2018-07-30T19:07:11.505Z",
                        "nadir": "2018-07-30T07:07:11.505Z",
                        "sunrise": "2018-07-30T11:57:49.750Z",
                        "sunset": "2018-07-31T02:16:33.259Z",
                        "sunriseEnd": "2018-07-30T12:00:53.253Z",
                        "sunsetStart": "2018-07-31T02:13:29.756Z",
                        "dawn": "2018-07-30T11:27:27.911Z",
                        "dusk": "2018-07-31T02:46:55.099Z",
                        "nauticalDawn": "2018-07-30T10:50:01.621Z",
                        "nauticalDusk": "2018-07-31T03:24:21.389Z",
                        "nightEnd": "2018-07-30T10:08:47.846Z",
                        "night": "2018-07-31T04:05:35.163Z",
                        "goldenHourEnd": "2018-07-30T12:36:14.026Z",
                        "goldenHour": "2018-07-31T01:38:08.983Z",
                    },
                    "sun_position": {
                        "azimuth": 0.9567419441563509,
                        "altitude": 1.0235714275875594,
                    },
                },
            }
        }

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_uv_index_with_api_status_check_first(
    aresponses: ResponsesMockServer,
    api_status_response: dict[str, Any],
    uv_index_response: dict[str, Any],
) -> None:
    """Test successfully retrieving UV index info after confirming the API status.

    Args:
        aresponses: An aresponses server.
        api_status_response: An API response payload.
        uv_index_response: An API response payload.
    """
    aresponses.add(
        "api.openuv.io",
        "/api/v1/status",
        "get",
        response=aiohttp.web_response.json_response(api_status_response, status=200),
    )
    aresponses.add(
        "api.openuv.io",
        "/api/v1/uv",
        "get",
        response=aiohttp.web_response.json_response(uv_index_response, status=200),
    )
    aresponses.add(
        "api.openuv.io",
        "/api/v1/status",
        "get",
        response=aiohttp.web_response.json_response({"status": False}, status=200),
    )

    async with aiohttp.ClientSession() as session:
        client = Client(
            TEST_API_KEY,
            TEST_LATITUDE,
            TEST_LONGITUDE,
            altitude=TEST_ALTITUDE,
            session=session,
            check_status_before_request=True,
        )

        # Test getting the data with a successful API status check:
        data = await client.uv_index()
        assert data == {
            "result": {
                "uv": 8.2342,
                "uv_time": "2018-07-30T20:53:06.302Z",
                "uv_max": 10.3335,
                "uv_max_time": "2018-07-30T19:07:11.505Z",
                "ozone": 300.7,
                "ozone_time": "2018-07-30T18:07:04.466Z",
                "safe_exposure_time": {
                    "st1": 20,
                    "st2": 24,
                    "st3": 32,
                    "st4": 40,
                    "st5": 65,
                    "st6": 121,
                },
                "sun_info": {
                    "sun_times": {
                        "solarNoon": "2018-07-30T19:07:11.505Z",
                        "nadir": "2018-07-30T07:07:11.505Z",
                        "sunrise": "2018-07-30T11:57:49.750Z",
                        "sunset": "2018-07-31T02:16:33.259Z",
                        "sunriseEnd": "2018-07-30T12:00:53.253Z",
                        "sunsetStart": "2018-07-31T02:13:29.756Z",
                        "dawn": "2018-07-30T11:27:27.911Z",
                        "dusk": "2018-07-31T02:46:55.099Z",
                        "nauticalDawn": "2018-07-30T10:50:01.621Z",
                        "nauticalDusk": "2018-07-31T03:24:21.389Z",
                        "nightEnd": "2018-07-30T10:08:47.846Z",
                        "night": "2018-07-31T04:05:35.163Z",
                        "goldenHourEnd": "2018-07-30T12:36:14.026Z",
                        "goldenHour": "2018-07-31T01:38:08.983Z",
                    },
                    "sun_position": {
                        "azimuth": 0.9567419441563509,
                        "altitude": 1.0235714275875594,
                    },
                },
            }
        }

        # Test raising when the API status check fails on a second attempt:
        with pytest.raises(ApiUnavailableError):
            data = await client.uv_index()

    aresponses.assert_plan_strictly_followed()
